# Daniel Zingaro, August 2011

# August 15 2011: don't crash on files with no valid questions
# August 15 2010: fix incompatibility with Iclicker 5.4.2 
# June 24 2010: added all_sessions.csv output of correctness for each question
#  in all sessions
# Process L*.csv files in date-order now
# April 2 2010: initial release

#Extract data from iclicker 5.2-5.4 .csv files
#Uses files in the current directory that match L*.csv


import glob, datetime

def make_date (clickstamp):
  '''Takes a string of the form yymmdd and returns a date string.'''
  
  year = '20' + clickstamp[0:2]
  month = clickstamp[2:4]
  day = clickstamp[4:6]
  if year.isdigit() and month.isdigit() and day.isdigit():
    datestamp = datetime.date (int(year), int(month), int(day))
  else:
    datestamp = 'none'
  return str(datestamp)
  
def file_cmp (fname1, fname2):
  date1 = fname1[1:7]
  date2 = fname2[1:7]
  return date1 < date2
      
def find_any (s, chars):  
  '''s is a list of strings; each string is empty or one character.
  chars is one or more characters.
  Return the number of characters of s found in chars.'''
  
  total = 0
  for ch in s:
    if ch != '' and ch in chars:
      total += 1
  return total

def get_average (percentages, loc, num_questions):
  if num_questions > 0:
    return sum([i[loc] for i in percentages]) / float(num_questions)
  else:
    return 0.0
    
    

def to_letter (num):
  '''Convert string of numbers in 1..5 to A..E.'''
  
  letters = 'XABCDE'
  if num.isdigit():
    nums = ''
    for number in num:
      nums = nums + letters[int(number)]
    return nums
  else:
    return num
  
def skip_lines (f, how_many):
  for i in range (how_many):
    f.readline()
    
def get_question_indices (s):
  '''Find all columns in csv that contain responses to a question.'''
    
  indices = []
  lst = s.split(',')
  for i in range(len(lst)):
    if lst[i].strip('"').lower().startswith ('question '):      
      indices.append (i)
  even_num = len(indices) % 2 == 0
  #For PI, we expect 2*num_ConcepTests
  if not even_num: print 'Warning: skipping file with ' + str(len(indices)) + ' questions'
  return indices, even_num
      
def get_correct (s, indices):  
  '''s is a line from the csv containing responses.
  Extract them based on the index of each response.'''
  
  correct = []
  lst = s.split(',')
  for i in indices:
    ans = lst[i].strip()
    if ans == '0': print 'Warning: question without correct answer'
    correct.append (to_letter(ans))
  return correct

def get_responses (f, indices):
  '''Return a dict whose keys are clicker IDs and whose values are
  lists of their respective responses.'''
  
  students = {}
  for line in f:	
    lst = line.split (',')
    cur_stu = lst[0]
    answers = get_correct (line, indices)
    students[cur_stu] = answers
  return students

def filter_singles (resp):  
  '''Eliminate votes existing in only one of the two votes for a ConcepTest.'''  
  
  for s in resp.values():
    for i in range(0, len(s), 2):
      if not s[i] or not s[i+1]:
        s[i] = ''
        s[i+1] = ''
  return resp

def transpose_responses (resp):
  '''convert from rows of students to rows of questions.'''
  
  trans = []
  for j in range(len(resp[0])):
    answers = []
    for i in range(len(resp)):
      answers.append (resp[i][j])
    trans.append (answers)
  return trans  

def get_percentages (responses, correct):
  '''Return list of (solo, group) correctness for each question.'''
  
  percentages = []
  for i in range(0, len(responses), 2):
    solo = responses[i]
    group = responses[i+1]
    solo_correct = float(find_any (solo, correct[i]))
    group_correct = float(find_any (group, correct[i]))
    num_votes = len(solo) - solo.count('')
    if num_votes > 0:
      percentages.append ((solo_correct / num_votes, group_correct / num_votes))
    else:
      percentages.append ((0, 0))
  return percentages

def get_normalized (percentages):
  '''Calculate NG based on solo/group correctness.'''
  
  ng = []
  for (solo, group) in percentages:
    if group > solo or solo == 0:
      new_ng = (group - solo) / (1 - solo)
    else:
      new_ng = (group - solo) / (solo)    
    ng.append (new_ng)
  return ng

def get_mazur (percentages, ind, iles):
  '''Divide questions into bins based on solo/group correctness.'''
  
  bins = [0] * len (iles)
  for i in range (len(percentages)):
    cur = percentages[i][ind]
    j = 0    
    while cur > iles[j]:
      j += 1
    bins[j] += 1
  return bins

def print_stats (percentages, num_lectures):
  ng = get_normalized (percentages)
  iles = [0.35, 0.70, 1]
  mazur = get_mazur (percentages, 0, iles)
  num_questions = len (percentages)
  average_per_lecture = float(num_questions) / num_lectures
  print 'Total questions: ', num_questions
  print 'Average number of questions per lecture: ', average_per_lecture
  average_solo = get_average (percentages, 0, num_questions)
  average_group = get_average (percentages, 1, num_questions)
  print 'Average correctness for solo vote: ', average_solo
  print 'Average correctness for group vote: ', average_group
  if len(ng) > 0:  
    average_ng = sum (ng) / len (ng)
  else:
    average_ng = 0
  print 'Average normalized gain: ', average_ng
  if num_questions == 0: num_questions = 1
  print 'Splitting on ', [str(i) for i in iles], ' : ', [float(i) / num_questions for i in mazur]

def dump_pairs (fname, pairs, difference):
  '''Dump pairs of data.'''
  
  f = open (fname, 'w')
  for (pre, post) in pairs:
    if difference:
      f.write (str(pre) + ' ' + str(post - pre) + '\n')
    else:
      f.write (str(pre) + ' ' + str(post) + '\n')

def dump_lists (fname, lists, header, sep=','):
  '''Dump lists of data.'''
  
  f = open (fname, 'w')
  f.write (header + '\n')
  for line in lists:
    text = sep.join ([str(s) for s in line])  
    f.write (text + '\n')

def get_abcde (responses, correct, fname):
  '''Return list where each tuple is of the following format:
  (date, q_number, percent_correct, num_a, num_b, num_c, num_d, num_e)'''
  
  tuples = []
  for i in range(0, len(responses)):    
    cur = responses[i]    
    num_correct = float(find_any (cur, correct[i]))    
    num_votes = len(cur) - cur.count('')
    five = []
    for j in 'ABCDE':
      five.append (float(find_any (cur, str(j))))
    file_date = make_date(fname[1:7])
    if num_votes > 0:
      perc = round(num_correct / num_votes, 2)
    else:
      perc = 0  
    tup = [file_date, str(i + 1), perc]+ five
    tuples.append (tup)
  return tuples

    
abcde = []
percentages = []
file_list = glob.glob ('L*.csv')
file_list.sort(file_cmp)
num_lectures = len(file_list)

for cur_file in file_list:
  print 'Processing: ' + cur_file
  f = open (cur_file)
  f.readline()
  index_line = f.readline()
  indices, cont = get_question_indices (index_line)
  if cont: # even number of questions?
    skip_lines (f, 2)
    correct_line = f.readline()
    correct = get_correct (correct_line, indices)
    skip_lines (f, 5)
    responses = get_responses (f, indices)
    trans_responses = transpose_responses (responses.values())    
    abcde.extend (get_abcde (trans_responses, correct, cur_file))
    responses = filter_singles (responses)
    trans_responses = transpose_responses (responses.values())    
    percentages.extend (get_percentages (trans_responses, correct))
print_stats (percentages, num_lectures)
percentages.sort ()
percentages = [[round(num1, 2), round(num2, 2)] for num1, num2 in percentages]
dump_pairs ('histo', percentages, True)
header = 'Date, Question, Percent Correct, A Count, B Count, C Count, D Count, E Count'
dump_lists ('all_sessions.csv', abcde, header)
percentages = [i for i in percentages if i[0] >= 0.35 and i[0] <= 0.7]
print '***Restricting to solo vote correctness in (35%, 70%)***'
print_stats (percentages, num_lectures)
